#ifndef _BLUEACLREQUESTHANDLERINTERFACE_H_
#define _BLUEACLREQUESTHANDLERINTERFACE_H_
class BlueAclRequestHandlerInterface
{

public:
        //Connection states
        enum  CONNECTION_STATE {	
		BTC_CLOSED = 0, // channel is closed ( all the calls return ERR_STATE )
		BTC_CLOSING, //channel is being closed ( can not be reverted, all the calls return ERR_STATE )
		BTC_OPEN, // channel is open and ready to exchange data through RFCOMM ( Send() / Receive() )
		BTC_OPEN_SETUP, //an external application is being setup over the channel
		BTC_OPEN_READY, //an external application is successfully setup
		BTC_OPENING, // channel is being opened ( using Open() ), Send() requests are queued, Receive() return ERR_STATE
		BTC_LISTENING // channel is listening for incoming connection, when a connection is received goes to OPEN state
	};

	//Error codes
	enum ERROR_CODE {
		ERR_OK = 0, // memset to 0 can be used for initializing
		ERR_STATE, // reported if an operation was executed in the wrong state
		ERR_FAILED, //operation failed
		ERR_CONNECTION //connection with buetooth device is lost
	};

	//Bluetooth profiles
	enum PROFILES {
		PROF_HF,
		PROF_DUN
	};

	// Construct a bluetooth channel
	// Return: a ERR_OK on success or ERR_FAILED if can't setup UNIX socket connection with BlueAclServer
	virtual int	Construct() = 0;

	// Init a bluetooth channel
	//	aBdaddr - device address
	//	aChannel - channel number
        //      aProfile (bt profile type ID, so deamon can do extra things like start/stop ppp for DUN)
	//	aAutoReconnect - 0=do not reconnect 1=reconnect if connection is lost
	// Return: a handle to the open channel or -1 if opening failed
	virtual int	Init( char* aBdaddr, int aChannel, int aProfile, int aAutoReconnect = 0 ) = 0;


	// Start additional use of this channel. For instance, when invoked on a DUN channel, the ppp deamon will be started.
        // On HF, it starts SCO, etcetera.
	//	aHandle - channel handle
	// Return: see ERR_CODE
	virtual int	StartUsingChannel( int aHandle, char* aData ) = 0;


	// Stop additional use of this channel. For instance, when invoked on a DUN channel, the ppp deamon will be stopped.
        // On HF, it stops SCO, etcetera.
	//	aHandle - channel handle
	// Return: see ERR_CODE
	virtual int	StopUsingChannel( int aHandle) = 0;


	// Close the channel (deamon does nopt listen to the incoming connections on the channel, handle is invalidated)
	//	aHandle - channel handle
	// Return: see ERR_CODE
	virtual int	Close( int aHandle, int aForced ) = 0;


	// Release channel (that is close the connection, go to listening mode, handle is still valid)
	//	aHandle - channel handle
	// Return: see ERR_CODE
	virtual int	Release( int aHandle ) = 0;

	// Send data over the channel (in needed the channel is opened)
	//	aHandle - channel handle
	//	aBuffer - pointer to the buffer to be sent (contains a ZT string)
	// Return: see ERROR_CODE
	virtual int	Send(int aHandle, char* aBuffer, int aSize) = 0;

	// Receive data from the bluetooth channel (if needed the channel is opened)
	//	aHandle - channel handle
	//	aBuffer - pointer to the receive buffe

	//	aSize - receive buffer size (a ZT string is returned)
	// Return: see ERROR_CODE
	virtual int	Receive(int aHandle, char* aBuffer, int* aSize) = 0;


	// Returns connection state
	//	aHandle - channel handle
	// Return: state as specified by CONNECTION_STATE or -1 in case of error
	virtual int	GetState( int aHandle ) = 0;


	// Returns the result of the latest asynch operation on the channel (listen, send)
	//	aHandle - channel handle
	// Return: error code ERROR_CODE
	virtual int	GetError( int aHandle ) = 0;

	// Set/Reset auto reconnect feature
	//	aHandle - channel handle
	//	aAutoReconnect - true /false == enabled/disabled
	// Return: -
	virtual void	SetAutoReconnect( int aHandle, int aAutoReconnect ) = 0;

	// Returns autoreconnect status
	//	aHandle - channel handle
	// Return: autoreconnect status
	virtual int	GetAutoReconnect( int aHandle ) = 0;

	// Returns is selected profile is active
	//	aProfile - profile identifier
	// Return: 0 - inactive, ~0 - active
	virtual int	IsProfileActive( int aProfile ) = 0;

};
#endif //_BLUEACLREQUESTHANDLERINTERFACE_H_
