#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <unistd.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "blueaclrequesthandler.h"

//if you want to have asserts on 0 -> 1
#if 0
  #include <assert.h>
  #define BARH_ASSERT(x) do{ assert(x); } while(0)
#else
  #define BARH_ASSERT(x) do { } while (0)
#endif

//SOUND_DIR
#define noNEW_JAS_DEAMON

//#define DEBUG_ACLREQUESTHANDLER

#ifdef DEBUG_ACLREQUESTHANDLER
  #define LP_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LP_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LP_DEBUG(msg...)     do { } while (0)
  #define LP_DEBUGT(t, msg...) do { } while (0)
#endif

/*****************************************************************************



			Class method implementation


*****************************************************************************/

BlueAclRequestHandler::BlueAclRequestHandler()
{
	int i, r;

	r = pthread_mutex_init(&handlerMutex, NULL);
	BARH_ASSERT(0 == r);

	// Mark all the channels as closed
	for(i = 0; i<MAX_HANDLES; i++)
	{
		handlerConnection[ i ].state = BTC_CLOSED;
		handlerConnection[ i ].profile = -1;
	}

	// initialise PPP
	pppPid = -1;
	pppHandle = -1;
	pppRemove = 0;
	// Thread handling daemon-wide tasks
	systemPool.Initialize( 0 );
	systemPool.Add(&SystemThread, this, RequestPool::RP_ADD_TAIL);
}

int BlueAclRequestHandler::Init( char* aAddr, int aChannel, int aProfile, int aAutoReconnect  )
{
	LP_DEBUG("BlueAclRequestHandler::Init(%s, %d, %d).\n", aAddr, aChannel, aProfile);

	int r, handle;
	ChannelInfo* pInfo;

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);
	
	handle = GetHandle(aAddr, aChannel);

	if ( handle < 0 )
	{
		handle = AssignHandle();
		handlerConnection[ handle ].BtChannel.Initialize( aAddr, aChannel );
		handlerConnection[ handle ].requestPool.Initialize( 0 );
		handlerConnection[ handle ].profile = aProfile;
		handlerConnection[ handle ].refCount = 1;
		handlerConnection[ handle ].useRefCount = 0;
		handlerConnection[ handle ].autoReconnect = aAutoReconnect;
		handlerConnection[ handle ].reconnectRetry = RECONNECT_COUNT;
		handlerConnection[ handle ].error = ERR_OK;
#ifdef NEW_JAS_DEAMON
		if ( aProfile == PROF_HF )
		{
			if ( jas_hf_init(&(handlerConnection[ handle ].scoHandle), 0) < 0 )
			{
				LP_DEBUG("BlueAclRequestHandler::Init(): SCO initialization FAILED!\n");
			}
		}
#endif
		CondAssignState( handle, BTC_LISTENING );
		if ( aProfile == PROF_HF )
		{
			pInfo = (ChannelInfo*)SetRequest(RQ_LISTEN, handle, NULL, 0);
			handlerConnection[ handle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);
		}
	}
	else
	{
		handlerConnection[ handle ].refCount++;
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::Init. Returned %d.\n", handle);

	return handle;
}

int BlueAclRequestHandler::StartUsingChannel( int aHandle, char* aData )
{
	LP_DEBUG("BlueAclRequestHandler::StartUsingChannel(%d), profile %d, data [%x]\n", aHandle, handlerConnection[ aHandle ].profile, aData);

	int r, error;
	ChannelInfo* pInfo;

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( handlerConnection[ aHandle ].state != BTC_CLOSED && handlerConnection[ aHandle ].state != BTC_CLOSING )
	{
		switch( handlerConnection[ aHandle ].profile )
		{
		case PROF_DUN:
			if ( ( pppPid < 0 ) || ( pppHandle < 0 ) || ( pppHandle == aHandle ) )
			{	
				if ( !handlerConnection[ aHandle ].useRefCount || ( handlerConnection[ aHandle].state == BTC_LISTENING ) )
				{
					CondAssignState( aHandle, BTC_OPEN_SETUP );
					pInfo = (ChannelInfo*)SetRequest(RQ_START_DUN, aHandle, aData, strlen( aData ) + 1 );
					handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);
				}
				handlerConnection[ aHandle ].useRefCount++;
				error = ERR_OK;
			}
			else
			{
				error = ERR_FAILED;
			}
			break;
		case PROF_HF:
			if ( !handlerConnection[ aHandle ].useRefCount )
			{
#ifdef NEW_JAS_DEAMON
				if ( jas_hf_listen( &(handlerConnection[ aHandle ].scoHandle), handlerConnection[ aHandle ].BtChannel.GetAddress() ) < 0 )
				{
					LP_DEBUG("BlueAclRequestHandler::DoStartHF(): Can not listen to the SCO channel!\n");
					error = ERR_FAILED;
				}
				else
#endif
				{
					handlerConnection[ aHandle ].useRefCount++;
					error = ERR_OK;
				}
			}
			break;
		default:
			BARH_ASSERT( false );
		}
	}
	else
	{
		error = ERR_STATE;
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::StartUsingChannel. Returned %d.\n", error);

	return error;
}

int BlueAclRequestHandler::StopUsingChannel( int aHandle )
{
	LP_DEBUG("BlueAclRequestHandler::StopUsingChannel(%d), profile %d\n", aHandle, handlerConnection[ aHandle ].profile);

	int r;
	int error;

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( !handlerConnection[ aHandle ].useRefCount )
	{
		r = pthread_mutex_unlock(&handlerMutex);
		BARH_ASSERT(0 == r);
		LP_DEBUG("BlueAclRequestHandler::StopUsingChannel: reference count: too many dereferences!\n");
		return ERR_STATE;
	}

	if ( --handlerConnection[ aHandle ].useRefCount )
	{
		LP_DEBUG("BlueAclRequestHandler::StopUsingChannel. Dereferenced count = %d.\n", handlerConnection[ aHandle ].useRefCount);
		error = ERR_OK;
	}
	else
	{
		LP_DEBUG("BlueAclRequestHandler::StopUsingChannel. Stopping...\n");
		switch( handlerConnection[ aHandle ].profile )
		{
		case PROF_DUN:
			CondAssignState( aHandle, BTC_LISTENING );
			pppHandle = -1;
			pppRemove = 1;
			error = ERR_OK;
			break;
		case PROF_HF:
#ifdef NEW_JAS_DEAMON
			if ( jas_hf_close( &(handlerConnection[ aHandle ].scoHandle) ) < 0 )
			{
				LP_DEBUG("BlueAclRequestHandler::DoStopHF(): closing SCO connection failed!\n");
				error = ERR_FAILED;
			}
			else
#endif
			{
				error = ERR_OK;
			}
			break;
		default:
			BARH_ASSERT( false );
		}
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::StopUsingChannel. Returned %d.\n", error);

	return error;
}

int BlueAclRequestHandler::Close( int aHandle, int aForced )
{
	LP_DEBUG("BlueAclRequestHandler::Close(%d, %d).\n", aHandle, aForced);

	int		r;
	int		error = ERR_OK;
	ChannelInfo*	pInfo;

	BARH_ASSERT( aHandle >= 0 );

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( handlerConnection[ aHandle].state != BTC_CLOSED  && handlerConnection[ aHandle].state != BTC_CLOSING  )
	{
		handlerConnection[ aHandle ].refCount--;
		BARH_ASSERT( handlerConnection[ aHandle ].refCount >= 0 );

		if ( handlerConnection[ aHandle ].refCount == 0 )
		{

			if( ( handlerConnection[ aHandle ].profile != PROF_DUN ) && ( ( handlerConnection[ aHandle ].state == BTC_LISTENING ) || aForced ) )
			{
				//force exit
				handlerConnection[ aHandle ].requestPool.Destroy(RequestPool::RP_DESTROY_NOW);
				handlerConnection[ aHandle ].requestPool.Initialize(0);
			}

			CondAssignState( aHandle, BTC_CLOSING );

			//dereference and close external app
			if ( handlerConnection[ aHandle ].useRefCount )
			{
				LP_DEBUG("BlueAclRequestHandler::Close: WARNING! The channel use dereference forced!\n");
				handlerConnection[ aHandle ].useRefCount = 0;
				switch( handlerConnection[ aHandle ].profile )
				{
				case PROF_DUN:
					pppHandle = -1;
					pppRemove = 1;
					break;
				case PROF_HF:
#ifdef NEW_JAS_DEAMON
					if ( jas_hf_close( &(handlerConnection[ aHandle ].scoHandle) ) < 0 )
						LP_DEBUG("BlueAclRequestHandler::Close(): closing SCO connection failed!\n");
#endif
					break;
				default:
					BARH_ASSERT( false );
				}
			}
			
			//start closing the channel
			pInfo = (ChannelInfo*)SetRequest(RQ_CLOSE, aHandle, NULL, 0);
			handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);
			//and then destroy the pool
			handlerConnection[ aHandle ].requestPool.Destroy(RequestPool::RP_DESTROY_ASYNC);
		}
		error = ERR_OK;
	}
	else
	{
		error = ERR_STATE;
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::Close. Returned %d.\n", error);
	return error;
}

int BlueAclRequestHandler::Release( int aHandle )
{
	LP_DEBUG("BlueAclRequestHandler::Release(%d).\n", aHandle);

	int		r;
	int		error;
	ChannelInfo*	pInfo;


	BARH_ASSERT( aHandle >= 0 );

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if( handlerConnection[ aHandle].state == BTC_OPEN )
	{
		//switch to listening
		CondAssignState( aHandle, BTC_LISTENING );
		
		//start releasing
		pInfo = (ChannelInfo*)SetRequest(RQ_RELEASE, aHandle, NULL, 0);
		handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);

		error = ERR_OK;

	}
	else
	{
		error = ERR_STATE;
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::Release. Returned %d.\n", error);
	return error;
}

int BlueAclRequestHandler::Send(int aHandle, char* aBuffer, int aSize)
{
#ifdef nDEBUG_ACLREQUESTHANDLER
	BARH_ASSERT( aSize <= 4096 );
	char	dbuf[4096];
	strncpy( dbuf, aBuffer, aSize );
	dbuf[aSize]='\0';
	LP_DEBUG("BlueAclRequestHandler::Send(%d, [%s], %d).\n", aHandle, dbuf, aSize);
#endif

	int		r;
	int		result;
	ChannelInfo*	pInfo;

	BARH_ASSERT( aHandle >= 0 );

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( ( handlerConnection[ aHandle ].state == BTC_LISTENING ) ||
		 ( handlerConnection[ aHandle ].state == BTC_OPEN ) ||
		 ( handlerConnection[ aHandle ].state == BTC_OPENING ) )
	{

		//add opening thread an sending thread to the pool
		if ( handlerConnection[ aHandle ].state == BTC_LISTENING )
		{
			CondAssignState( aHandle, BTC_OPENING );
			//force exit
			handlerConnection[ aHandle ].requestPool.Destroy(RequestPool::RP_DESTROY_NOW);
			handlerConnection[ aHandle ].requestPool.Initialize(0);
			

			//open the channel
			pInfo = (ChannelInfo*)SetRequest(RQ_OPEN, aHandle, NULL, 0);
			handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);

		}
		//start sending the data
		pInfo = (ChannelInfo*)SetRequest(RQ_SEND, aHandle, aBuffer, aSize);
		handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);

		result = ERR_OK;
	}
	else
	{
		result = ERR_STATE;
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

//	LP_DEBUG("BlueAclRequestHandler::Send. Returned %d.\n", result);

	return result;

}

int BlueAclRequestHandler::Receive(int aHandle, char* aBuffer, int* aSize)
{
	int 	r;
	int 	error;
	ChannelInfo*	pInfo;

//	LP_DEBUG("BlueAclRequestHandler::Receive(%d, %x, %d).\n", aHandle, aBuffer, *aSize);

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if( handlerConnection[ aHandle].state == BTC_OPEN )
	{
		r = handlerConnection[ aHandle ].BtChannel.Receive(aBuffer, aSize);
		if( ( r < 0 ) && handlerConnection[ aHandle ].autoReconnect )
		{
			LP_DEBUG("BlueAclRequestHandler::Receive() connection is lost!\n");

			CondAssignState( aHandle, BTC_OPENING );

			//start reopening
			pInfo = (ChannelInfo*)SetRequest(RQ_OPEN, aHandle, NULL, 0);
			handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_HEAD);

			handlerConnection[ aHandle ].error = ERR_CONNECTION;
			error = ERR_CONNECTION;
		}
		else if ( ( r < 0 ) && !handlerConnection[ aHandle ].autoReconnect )
		{

			handlerConnection[ aHandle ].requestPool.Destroy(RequestPool::RP_DESTROY_NOW);
			handlerConnection[ aHandle ].requestPool.Initialize(0);

			CondAssignState( aHandle, BTC_LISTENING );
			//start listening
			pInfo = (ChannelInfo*)SetRequest(RQ_LISTEN, aHandle, NULL, 0);
			handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_HEAD);

			handlerConnection[ aHandle ].error = ERR_CONNECTION;
			error = ERR_CONNECTION;
			
		}
		else
		{
			handlerConnection[ aHandle ].error = ERR_OK;
			error = ERR_OK;
		}
	}
	else
	{
		*aSize = 0;
		error = ERR_STATE;	
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

#ifdef DEBUG_ACLREQUESTHANDLER
//	BARH_ASSERT( *aSize <= 4096 );
//	char	dbuf[4096];
//	strncpy( dbuf, aBuffer, *aSize );
//	dbuf[*aSize] = '\0';
//	LP_DEBUG("BlueAclRequestHandler::Receive. Returned error %d, buffer [%s].\n", error, dbuf );
#endif
	return error;
}


int BlueAclRequestHandler::GetState( int aHandle )
{
//	LP_DEBUG("BlueAclRequestHandler::GetState. Returned %d.\n", handlerConnection[ aHandle ].state);
	return handlerConnection[ aHandle ].state;
}


int BlueAclRequestHandler::GetError( int aHandle )
{
//	LP_DEBUG("BlueAclRequestHandler::GetError. Returned %d.\n", handlerConnection[ aHandle ].error);
	return handlerConnection[ aHandle ].error;
}

void BlueAclRequestHandler::SetAutoReconnect( int aHandle, int aAutoReconnect )
{
	int	r;

	LP_DEBUG("BlueAclRequestHandler::SetAutoReconnect(%d, %d).\n", aHandle, aAutoReconnect);
	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	handlerConnection[ aHandle ].autoReconnect = aAutoReconnect;

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);
	LP_DEBUG("BlueAclRequestHandler::SetAutoReconnect() returned.\n");
}

int BlueAclRequestHandler::GetAutoReconnect( int aHandle )
{
	LP_DEBUG("BlueAclRequestHandler::GetAutoReconnect(%d) returned %d.\n", aHandle, handlerConnection[ aHandle ].autoReconnect);
	return handlerConnection[ aHandle ].autoReconnect;
}

int BlueAclRequestHandler::IsProfileActive( int aProfile )
{
	int result = 0;

	switch( aProfile )
	{
	case PROF_DUN:
		if ( pppPid > 0 )
			result = 1;
		break;
	case PROF_HF:
		for ( int i = 0; i < MAX_HANDLES; i++ )
		{
			if ( ( handlerConnection[ i ].profile == PROF_HF ) &&  ( handlerConnection[ i ].state == BTC_OPEN ) )
			{
				result = 1;
				break;
			}
		}
		break;
	}

	return result;
}

/*****************************************************************************



				Thread routines


*****************************************************************************/
void BlueAclRequestHandler::DoListen( int aHandle )
{
	//  This is being executed in a separate thread!

//	BARH_ASSERT( handlerConnection[ aHandle ].state == BTC_LISTENING );

	int	r;
	int	result;

	LP_DEBUG("BlueAclRequestHandler::DoListen(). Listening on socket %s, channel %d !\n", handlerConnection[ aHandle ].BtChannel.GetAddress(),
											      ( handlerConnection[ aHandle ].profile == PROF_HF ) ?
													HF_PROFILE_CHANNEL :
													handlerConnection[ aHandle ].BtChannel.GetChannel());
	//close previously open connections
	if ( handlerConnection[ aHandle ].BtChannel.Close() < 0 )
	{
		LP_DEBUG("BlueAclRequestHandler::DoListen(%d). BtChannel.Close() FAILED!\n", aHandle);
	}
	if ( handlerConnection[ aHandle ].profile == PROF_HF )
		result = handlerConnection[ aHandle ].BtChannel.Listen(HF_PROFILE_CHANNEL);
	else
		result = handlerConnection[ aHandle ].BtChannel.Listen(handlerConnection[ aHandle ].BtChannel.GetChannel());

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( result < 0 )
	{
		LP_DEBUG("LISTEN(%d). FAILED!\n", aHandle);


		CondAssignState( aHandle, BTC_CLOSED );
		handlerConnection[ aHandle ].error = ERR_OK;

		handlerConnection[ aHandle ].requestPool.Destroy(RequestPool::RP_DESTROY_ASYNC);

	}
	else
	{
		LP_DEBUG("LISTEN(%d). SUCCEDED!\n", aHandle);

		CondAssignState( aHandle, BTC_OPEN );
		handlerConnection[ aHandle ].error = ERR_OK;

	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);
		
	LP_DEBUG("BlueAclRequestHandler::DoListen. Returned %d.\n", handlerConnection[ aHandle ].error);

}


void BlueAclRequestHandler::DoClose( int aHandle )
{
	
	BARH_ASSERT( handlerConnection[ aHandle ].state == BTC_CLOSING );

	int r, result, cnt;

	LP_DEBUG("BlueAclRequestHandler::DoClose(). Closing socket %s, channel %d !\n", handlerConnection[ aHandle ].BtChannel.GetAddress(),
											handlerConnection[ aHandle ].BtChannel.GetChannel());

	// wait for the previous PPP session to finish
	cnt = PPP_SHUTDOWN_CHECK_COUNT;
	while( ( pppPid > 0 ) && --cnt )
	{
		LP_DEBUG("BlueAclRequestHandler::DoClose(): Waiting for the previous PPP session to shutdown...\n");
		sleep( PPP_STARTUP_CHECK_INTERVAL );
	}

	result = handlerConnection[ aHandle ].BtChannel.Close();
	
	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	CondAssignState( aHandle, BTC_CLOSED );

	if ( result < 0 )
	{
		LP_DEBUG("BlueAclRequestHandler::DoClose(%d). FAILED!\n", aHandle);
		handlerConnection[ aHandle ].error = ERR_FAILED;
	}

	handlerConnection[ aHandle ].error = ( result < 0 ) ? ERR_FAILED : ERR_OK;

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::DoClose. Returned %d.\n", handlerConnection[ aHandle ].error);

}

                                      	
void BlueAclRequestHandler::DoRelease( int aHandle )
{
	//  This is being executed in a separate thread!

	BARH_ASSERT( handlerConnection[ aHandle ].state == BTC_LISTENING );

	int	r;
	int	result;

	LP_DEBUG("BlueAclRequestHandler::DoRelease(). Listening on socket %s, channel %d !\n", handlerConnection[ aHandle ].BtChannel.GetAddress(),
											       ( handlerConnection[ aHandle ].profile == PROF_HF ) ? 
													HF_PROFILE_CHANNEL:
													handlerConnection[ aHandle ].BtChannel.GetChannel());

	//close previously open connections
	if ( handlerConnection[ aHandle ].BtChannel.Close() < 0 )
	{
		LP_DEBUG("BlueAclRequestHandler::Release(%d). BtChannel.Close() FAILED!\n", aHandle);
	}

	if ( handlerConnection[ aHandle ].profile == PROF_HF )
		result = handlerConnection[ aHandle ].BtChannel.Listen(HF_PROFILE_CHANNEL);
	else
		result = handlerConnection[ aHandle ].BtChannel.Listen(handlerConnection[ aHandle ].BtChannel.GetChannel());

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( result < 0 )
	{
		LP_DEBUG("LISTEN(%d). FAILED!\n", aHandle);

		CondAssignState( aHandle, BTC_CLOSED );
		handlerConnection[ aHandle ].requestPool.Destroy(RequestPool::RP_DESTROY_NOW);

	}
	else
	{
		LP_DEBUG("LISTEN(%d). SUCCEDED!\n", aHandle);

		CondAssignState( aHandle, BTC_OPEN );
	}

	handlerConnection[ aHandle ].error = ERR_OK;

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::DoRelease. Returned %d.\n", handlerConnection[ aHandle ].error);

}

void BlueAclRequestHandler::DoSend( int aHandle, char* apData, int aSize )
{
	//  This is being executed in a separate thread!

	BARH_ASSERT( ( handlerConnection[ aHandle ].state == BTC_OPEN ) ||
		( handlerConnection[ aHandle ].state == BTC_OPENING ) ||
		( handlerConnection[ aHandle ].state == BTC_CLOSING ) ||
		( handlerConnection[ aHandle ].state == BTC_OPEN_SETUP ) );

#ifdef nDEBUG_ACLREQUESTHANDLER
	BARH_ASSERT( aSize <= 4096 );
	char	dbuf[4096];
	strncpy( dbuf, apData, aSize );
	dbuf[aSize]='\0';
	LP_DEBUG("BlueAclRequestHandler::DoSend(). Sending [%s] over socket %s, channel %d !\n", dbuf,
												 handlerConnection[ aHandle ].BtChannel.GetAddress(), 
												 handlerConnection[ aHandle ].BtChannel.GetChannel());
#endif
	int		sent;
	int		r;
	ChannelInfo*	pInfo;
	ChannelInfo*	pInfo1;

	sent = handlerConnection[ aHandle ].BtChannel.Send(apData, aSize);

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( sent < 0 )
	{
		LP_DEBUG("BlueAclRequestHandler::DoSend(). Initiating reopening!\n");

		CondAssignState( aHandle, BTC_OPENING );
		handlerConnection[ aHandle ].error = ERR_CONNECTION;

		pInfo = (ChannelInfo*)SetRequest(RQ_OPEN, aHandle, NULL, 0);
		pInfo1 = (ChannelInfo*)SetRequest(RQ_SEND, aHandle, apData, aSize);
		handlerConnection[ aHandle ].requestPool.Add2(&RequestThread, (void*)pInfo, &RequestThread, (void*)pInfo1, RequestPool::RP_ADD_HEAD);

	}
	else if(aSize != sent)
	{
		LP_DEBUG("Sending FAILED! Reported %d bytes, sent expected %d.\n", sent, aSize);

		handlerConnection[ aHandle ].error = ERR_FAILED;
	}
	else
	{
		handlerConnection[ aHandle ].error = ERR_OK;

	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

//	LP_DEBUG("BlueAclRequestHandler::DoSend. Returned %d.\n", handlerConnection[ aHandle ].error);

}
void BlueAclRequestHandler::DoOpen( int aHandle )
{
	//  This is being executed in a separate thread!

	BARH_ASSERT( ( handlerConnection[ aHandle ].state == BTC_OPENING ) || 
		( handlerConnection[ aHandle ].state == BTC_CLOSING ) || 
		( handlerConnection[ aHandle ].state == BTC_OPEN_SETUP ) );

	ChannelInfo*	pInfo;
	ChannelInfo*	pInfo1;

	int	r;
	int	result;

	LP_DEBUG("BlueAclRequestHandler::DoOpen(). Opening socket %s, channel %d, auto reconnect %d !\n", handlerConnection[ aHandle ].BtChannel.GetAddress(), 
													  handlerConnection[ aHandle ].BtChannel.GetChannel(), 
													  handlerConnection[ aHandle ].autoReconnect);
	//close previously open connections
	if ( handlerConnection[ aHandle ].BtChannel.Close() < 0 )
	{
		LP_DEBUG("Warning: BlueAclRequestHandler::DoOpen(%d). BtChannel.Close() FAILED!\n", aHandle);
	}

	result = handlerConnection[ aHandle ].BtChannel.Open();

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	if ( result < 0 )
	{
		LP_DEBUG("BlueAclRequestHandler::DoOpen(%d). FAILED!\n", aHandle);

		if ( handlerConnection[ aHandle ].autoReconnect || handlerConnection[ aHandle ].reconnectRetry )
		{
			if ( !handlerConnection[ aHandle ].autoReconnect ) 
				handlerConnection[ aHandle ].reconnectRetry--;

			LP_DEBUG("BlueAclRequestHandler::DoOpen(%d). Rescheduling OPEN!\n", aHandle);
			//start opening again (add delay)
			pInfo = (ChannelInfo*)SetRequest(RQ_SLEEP, aHandle, NULL, 0);
			pInfo1 = (ChannelInfo*)SetRequest(RQ_OPEN, aHandle, NULL, 0);
			handlerConnection[ aHandle ].requestPool.Add2(&RequestThread, (void*)pInfo, &RequestThread, (void*)pInfo1, RequestPool::RP_ADD_HEAD);
		}
		else
		{
			handlerConnection[ aHandle ].reconnectRetry = RECONNECT_COUNT;

			LP_DEBUG("BlueAclRequestHandler::DoOpen(%d). Switching to LISTENING!\n", aHandle);
			handlerConnection[ aHandle ].requestPool.Empty();
			//switch back to listening
			CondAssignState( aHandle, BTC_LISTENING );
			if (handlerConnection[ aHandle ].state != BTC_CLOSING)
			{
				//then start listening
				pInfo = (ChannelInfo*)SetRequest(RQ_LISTEN, aHandle, NULL, 0);
				handlerConnection[ aHandle ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_HEAD);
			}

		}
	}
	else
	{
		if ( handlerConnection[ aHandle ].state != BTC_OPEN_SETUP ) CondAssignState( aHandle, BTC_OPEN );
		handlerConnection[ aHandle ].reconnectRetry = 0;
	}

	
	handlerConnection[ aHandle ].error = ( result < 0 ) ? ERR_CONNECTION : ERR_OK;

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::DoOpen. Returned %d.\n", handlerConnection[ aHandle ].error);

}

void BlueAclRequestHandler::DoStartDUN( int aHandle, char* apData )
{
	int r, pid, st, cnt, status;
	ChannelInfo* pInfo;
	struct stat s;
	bool pppup = false, dnsdown=true;
	

	LP_DEBUG("BlueAclRequestHandler::DoStartDUN(%d, %s).\n", aHandle, (char*)apData);


	// wait for the previous PPP session to finish
	cnt = PPP_SHUTDOWN_CHECK_COUNT;
	while( ( pppPid > 0 ) && --cnt )
	{
		LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): Waiting for the previous PPP session to shutdown...\n");
		sleep( PPP_STARTUP_CHECK_INTERVAL );
	}

	pppRemove = 0;

	// create TTY to be used by PPP
	if ( handlerConnection[ aHandle ].BtChannel.CreateTTY( aHandle ) >= 0 || !cnt )
	{
		LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): Using TTY %s . \n", handlerConnection[ aHandle ].BtChannel.GetTTY());
	
		//prepare command line arguments
		char*	argv[6];
		int	n = 0;

		argv[n++] = PPP_CMDLINE;
		argv[n++] = handlerConnection[ aHandle ].BtChannel.GetTTY();
		argv[n++] = PPP_BAUDRATE;
		if ( strlen(apData) )
		{
			argv[n++] = "call";
			argv[n++] = (char*)apData;
		}
		argv[n++] = NULL;
		//execute
		LP_DEBUG( "Executing: %s %s %s %s %s\n", argv[0], argv[1], argv[2], argv[3], argv[4] );

		pid = vfork();

		if ( pid > 0 )
		{
			//process id of pppd is 'pid'
			LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): PPPD PID = %d\n", pid);
		}                               
		else if ( pid == 0 )
		{
		    	execvp( PPP_CMDLINE, argv );
			LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): execvp(pppd) FAILED!\n");
			_exit(1);
		}
		else
			LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): vfork() FAILED!\n");

		if ( pid > 0 )
		{
			cnt = PPP_STARTUP_CHECK_COUNT;
			while ( !(pppup = !( stat("/var/run/ppp.up", &s) < 0 )) && !waitpid(pid, &status, WNOHANG ) && --cnt )
			{                     	
				LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): checking ppp connection...\n");
				sleep( PPP_STARTUP_CHECK_INTERVAL );
			}

			// ppp is up, now check if DNS is there too (or if startup was cancelled
			if ( pppup && !pppRemove )
			{
				// check if real connection is up
				cnt = PPP_DNS_CHECK_COUNT;
				do
				{
					LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): resolving DNS...\n");
					sleep(PPP_DNS_CHECK_INTERVAL);
				}
				while( ( gethostbyname(PPP_CHECK_HOST) == NULL ) && --cnt && !pppRemove);

				if ( cnt )
					dnsdown = false;
				else
					dnsdown = true;
			}
		}

		handlerConnection[ aHandle ].BtChannel.ReleaseTTY();
	}

	r = pthread_mutex_lock(&handlerMutex);
	BARH_ASSERT(0 == r);

	pppPid = pid;
	pppHandle = aHandle;

	if( !pppup || dnsdown || pppRemove )
	{

		LP_DEBUG( "BlueAclRequestHandler::DoStartDun(): PPP connection failed!\ni");

		if ( pppPid > 0 )
		{
			LP_DEBUG( "BlueAclRequestHandler::DoStartDun(): removing pppd!\ni");
			kill(pppPid, SIGTERM);
		}
		CondAssignState( aHandle, BTC_LISTENING );
		handlerConnection[ aHandle ].error = ERR_FAILED;
	}
	else
	{
		LP_DEBUG("BlueAclRequestHandler::DoStartDUN(): PPP connection established!\n");
		CondAssignState( aHandle, BTC_OPEN_READY );
		handlerConnection[ aHandle ].error = ERR_OK;
	}

	r = pthread_mutex_unlock(&handlerMutex);
	BARH_ASSERT(0 == r);

	LP_DEBUG("BlueAclRequestHandler::DoStartDUN. Returned %d.\n", handlerConnection[ aHandle ].error);

}


/*****************************************************************************



				Service routines


*****************************************************************************/

void BlueAclRequestHandler::CondAssignState( int aHandle, int aState )
{
	int	oldState = handlerConnection[ aHandle ].state;

	if ( handlerConnection[ aHandle ].state != BTC_CLOSING &&
		( handlerConnection[ aHandle].state != BTC_OPEN_SETUP || 
			( ( handlerConnection[ aHandle].state == BTC_OPEN_SETUP ) && ( ( aState == BTC_OPEN_READY ) || ( aState == BTC_CLOSING ) || ( aState == BTC_LISTENING ) ) ) ) ||
				( aState == BTC_CLOSED ) )
	{
		handlerConnection[ aHandle ].state = aState;
	}
	LP_DEBUG("CondAssignState(%d): %d -> %d\n", aHandle, oldState, handlerConnection[ aHandle ].state );
}

void* BlueAclRequestHandler::SetRequest(int aType, int aHandle, char* apData, int aSize )
{
	ChannelInfo* pInfo;

#ifdef nDEBUG_ACLREQUESTHANDLER
	BARH_ASSERT( aSize <= 4096 );
	char	dbuf[4096];
	strncpy( dbuf, apData, aSize );
	dbuf[aSize]='\0';
	LP_DEBUG("BlueAclRequestHandler::SetRequest(%d, %d, %s, %d)\n", aType, aHandle, dbuf, aSize);
#endif
	//this info will be passed to the processing thread
	pInfo = (ChannelInfo*)malloc( sizeof(ChannelInfo) + aSize ) ;
	BARH_ASSERT( pInfo );
	pInfo->requestType = aType;
	pInfo->pBlueAclRequestHandler = this;		
	pInfo->handle = aHandle;
	pInfo->bufferSize = aSize;
	if ( apData ) 
	{
		BARH_ASSERT( aSize );
		memcpy( pInfo->buffer, apData, aSize );
	}

//	LP_DEBUG("BlueAclRequestHandler::SetRequest(). Done.\n");

	return (void*)pInfo;
}
void BlueAclRequestHandler::RequestThread(void* aArgs)
{
	// static memeber called from the requestPool
	ChannelInfo* pInfo = static_cast<ChannelInfo*> (aArgs);

	switch(pInfo->requestType)
	{
	case RQ_LISTEN:
		pInfo->pBlueAclRequestHandler->DoListen( pInfo->handle );
		break;
	case RQ_OPEN:
		pInfo->pBlueAclRequestHandler->DoOpen( pInfo->handle );
		break;
	case RQ_SEND:
		pInfo->pBlueAclRequestHandler->DoSend( pInfo->handle, pInfo->buffer, pInfo->bufferSize );
		break;
	case RQ_CLOSE:
		pInfo->pBlueAclRequestHandler->DoClose( pInfo->handle );
		break;
	case RQ_RELEASE:
		pInfo->pBlueAclRequestHandler->DoRelease( pInfo->handle );
		break;
	case RQ_SLEEP:
		sleep(AUTORECONNECT_INTERVAL);
		break;
	case RQ_START_DUN:
		pInfo->pBlueAclRequestHandler->DoStartDUN( pInfo->handle, pInfo->buffer );
		break;
	default:
		BARH_ASSERT(false);
	}
}

int BlueAclRequestHandler::GetHandle(char* aAddr, int aChannel)
{
	int	i;
	int 	handle = -1;

	for(i=0; i<MAX_HANDLES; i++)
	{
		char*	Addr = handlerConnection[ i ].BtChannel.GetAddress();
		int	Channel = handlerConnection[ i ].BtChannel.GetChannel();

		if ( ( strcmp(aAddr, Addr) == 0 ) && ( aChannel == Channel ) && ( handlerConnection[ i ].state != BTC_CLOSED ) && ( handlerConnection[ i ].state != BTC_CLOSING ) )
		{
			handle = i;
			break;
		}
	}

	return handle;
}

int BlueAclRequestHandler::AssignHandle()
{

	int	i;
	int 	handle = -1;

	for(i=0; i<MAX_HANDLES; i++)
	{
		if ( handlerConnection[ i ].state == BTC_CLOSED )
		{
			handle = i;
			break;
		}
	}

	return handle;
}

/*****************************************************************************



				Awaker process


*****************************************************************************/
void BlueAclRequestHandler::DoSystemThread()
{
	int r, reconnectTimeout = AUTORECONNECT_INTERVAL, status;
	ChannelInfo*	pInfo;
	struct stat s;
	
	while(1)
	{
		sleep( 1 ); //granularity is a second

		r = pthread_mutex_lock(&handlerMutex);
		BARH_ASSERT(0 == r);

		reconnectTimeout--;
		// auto reconnect feature
		if ( !reconnectTimeout )
		{
			for ( int i = 0; i < MAX_HANDLES; i++)
			{
				if ( handlerConnection[ i ].autoReconnect && ( handlerConnection[ i ].state == BTC_LISTENING ) )
				{
					LP_DEBUG("BlueAclRequestHandler::DoSystemThread. Opening the channel on handle %d...\n", i);
					//force exit
					handlerConnection[ i ].requestPool.Destroy(RequestPool::RP_DESTROY_NOW);
					handlerConnection[ i ].requestPool.Initialize(0);
	
					//start opening channel
					CondAssignState( i, BTC_OPENING );

					//open the channel
					pInfo = (ChannelInfo*)SetRequest(RQ_OPEN, i, NULL, 0);
					handlerConnection[ i ].requestPool.Add(&RequestThread, (void*)pInfo, RequestPool::RP_ADD_TAIL);
				}
			}
			reconnectTimeout = AUTORECONNECT_INTERVAL;
		}

		// ppp timeout feature
		if ( pppRemove )
		{
			if ( pppPid > 0 )
			{
				LP_DEBUG("BlueAclRequestHandler::DoSystemThread: removing PPP!\n");
				kill( pppPid, SIGTERM );
				pppRemove = 0;
			}
		}

		// ppp down detection
		if ( pppPid > 0 )
		{
			if ( stat("/var/run/ppp0.pid", &s) < 0 )
			{
				LP_DEBUG("BlueAclRequestHandler::DoSystemThread: PPP dead!\n");
				pppPid = -1;
				if ( pppHandle >= 0 )
					CondAssignState( pppHandle, BTC_LISTENING );
			}
		}

		r = pthread_mutex_unlock(&handlerMutex);
		BARH_ASSERT(0 == r);
	}
}
void BlueAclRequestHandler::SystemThread(void* aArgs)
{

	(static_cast<BlueAclRequestHandler*>(aArgs))->DoSystemThread();
}
