#ifndef _BLUEACLREQUESTHANDLER_H_
#define _BLUEACLREQUESTHANDLER_H_
//
// Bluetooth ACL request handler header file
//
#include <pthread.h>
#include "BlueAclRequestHandlerInterface.h"
#include "requestpool.h"
#include "bluetoothchannel.h"

// System settings
#define MAX_HANDLES			30 //Maximum number of connections possible (=max number of RFCOMM channels)
#define MAX_STATE_LEN			256 //Maximum statename length
#define AUTORECONNECT_INTERVAL		5 //in seconds
#define RECONNECT_COUNT			2 //the number of connect retries if reconnect fails (autoreconnect disabled)

// PPP settings
#define PPP_CMDLINE			"/sbin/pppd"
#define PPP_BAUDRATE			"115200"
#define	PPP_STARTUP_CHECK_INTERVAL	3 // interval for checking ppp connection (seconds)
#define	PPP_SHUTDOWN_CHECK_INTERVAL	3 // interval for checking ppp connection (seconds)
#define	PPP_STARTUP_CHECK_COUNT		10 // interval for checking ppp connection (seconds)
#define	PPP_SHUTDOWN_CHECK_COUNT	10 // interval for checking ppp connection (seconds)
#define PPP_SHUTDOWN_INTERVAL		0 //the interval after which PPP will be shut down if inactive
#define PPP_DNS_CHECK_COUNT		3 // number of checks for DNS availability
#define PPP_DNS_CHECK_INTERVAL		3 // interval for checking DNS availability
#define PPP_CHECK_HOST			"t.tt1.nl"

// HF channel
#define HF_PROFILE_CHANNEL		3 // channel reported by TTG sdp server

class BlueAclRequestHandler : public BlueAclRequestHandlerInterface
{

public:


	BlueAclRequestHandler();
	int Construct() { return ERR_OK; }
	int Init( char* aBdaddr, int aChannel, int aProfile, int aAutoReconnect = 0 );
	int StartUsingChannel( int aHandle, char* aData );
	int StopUsingChannel( int aHandle );
	int Release( int aHandle );
	int Close( int aHandle, int aForced );
	int Send(int aHandle, char* aBuffer, int aSize);
	int Receive(int aHandle, char* aBuffer, int* aSize);
	int GetState( int aHandle );
	int GetError( int aHandle );
	void SetAutoReconnect( int aHandle, int aAutoReconnect );
	int GetAutoReconnect( int aHandle );
	int IsProfileActive( int aProfile );

private:

	int	GetHandle(char* aAddr, int aChannel);
	int	AssignHandle();
	void	CondAssignState( int aHandle, int aState );

	enum
	{
		RQ_LISTEN,
		RQ_OPEN,
		RQ_SEND,
		RQ_CLOSE,
		RQ_RELEASE,
		RQ_SLEEP,
		RQ_START_DUN

	} REQUEST_TYPE;

	struct ChannelInfo
	{
		BlueAclRequestHandler*	pBlueAclRequestHandler;
		int			requestType;
		int			handle;
		int			bufferSize;
		char			buffer[];
	};
	//request processors
	void*		SetRequest(int aType, int aHandle, char* apData, int aSize);
	static void	RequestThread(void* aArgs);
	//request handlers
	void 		DoSend( int aHandle, char* apData, int aSize );
	void 		DoOpen( int aHandle );
	void 		DoListen( int aHandle );
	void 		DoClose( int aHandle );
	void 		DoRelease( int aHandle );
	void		DoStartDUN( int aHandle, char* apData );
	//awaker thread, reopens channels on timer
	static void	SystemThread(void* aArgs);
	void		DoSystemThread();

private:
        // Channel descriptor
        typedef struct
        {
	  BluetoothChannel	BtChannel;
	  int			profile; //bluetooth profile used for the given channel
          int			state; // CHANNEL_STATE
          int			refCount; //the number of applications referring the connection
	  int			useRefCount; //StartUsing increases, stop using decreases
	  int 			error; // ERROR_CODE
	  int			autoReconnect;
	  int			reconnectRetry;
	  RequestPool		requestPool; //serial pool of incoming requests
        } Connection;

	pthread_mutex_t	handlerMutex;
	Connection	handlerConnection[MAX_HANDLES];
	RequestPool	systemPool;

	// PPP process id used to monitor PPP connection, timeout - to inplement idle shutdown:
	int		pppPid;
	int		pppHandle;
	int		pppRemove;

};
#endif //_BLUEACLREQUESTHANDLER_H_
