#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/un.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <pthread.h>

#include "blueaclserver.h"
#include "blueaclrequesthandler.h"

//#define DEBUG_ACLSERVER

#ifdef DEBUG_ACLSERVER
  #define LP_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LP_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LP_DEBUG(msg...)     do { } while (0)
  #define LP_DEBUGT(t, msg...) do { } while (0)
#endif

//if you want to have asserts on 0 -> 1
#if 0
  #include <assert.h>
  #define BAS_ASSERT(x) do{ assert(x); } while(0)
#else
  #define BAS_ASSERT(x) do { } while (0)
#endif

// the maximal number of connections allowed to this server
#define MAX_CLIENTS	10

// asynchronoous request handler
static BlueAclRequestHandler barh;

// to keep server running
static int serverbusy = 1;

void requesthandler(int fd)
{
	int	request;

	read( fd, &request, sizeof(int) );

//	LP_DEBUG("BlueAclServer: received command %d on fd %d.\n", request, fd);

	switch( request )
	{
	case CMD_INIT:
		{
		char	bdaddr[BDADDR_SIZE];
		int	channel, profile, autoreconnect;
		
		read( fd, &bdaddr, BDADDR_SIZE );
		read( fd, &channel, sizeof(int));
		read( fd, &profile, sizeof(int));
		read( fd, &autoreconnect, sizeof(int));

		LP_DEBUG( "Calling Init(%s, %d, %d, %d)...\n", bdaddr, channel, profile, autoreconnect );

		int result = barh.Init( bdaddr, channel, profile, autoreconnect );

		LP_DEBUG( "Init() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_CLOSE:
		{
		int	handle, forced;

		read( fd, &handle, sizeof(int) );
		read( fd, &forced, sizeof(int) );

		LP_DEBUG( "Calling Close(%d, %d)...\n", handle, forced );

		int result = barh.Close( handle, forced );

		LP_DEBUG( "Close() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_RELEASE:
		{
		int	handle;

		read( fd, &handle, sizeof(int) );

		LP_DEBUG( "Calling Release(%d)...\n", handle );

		int result = barh.Release( handle );

		LP_DEBUG("Returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_SEND:
		{
		int	handle, size;
		char*	buf;

		read( fd, &handle, sizeof(int) );
		read( fd, &size, sizeof(int) );
		buf = (char*)malloc( size );
		read( fd, buf, size );

		LP_DEBUG("Calling Send(%d, %x, %d)...\n", handle, buf, size);

		int result = barh.Send( handle, buf, size );

		LP_DEBUG("Send() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		free( buf );

		break;
		}
	case CMD_RECEIVE:
		{
		int	handle, size;
		char*	buf;

		read( fd, &handle, sizeof(int) );
		read( fd, &size, sizeof(int) );
		buf = (char*)malloc( size );

//		LP_DEBUG("Calling Receive(%d, %d)...\n", handle, size );

		int result = barh.Receive( handle, buf, &size );

//		LP_DEBUG("Returned, size = %d, error = %d.\n", size, result);

		write( fd, &result, sizeof(int) );
		write( fd, &size, sizeof(int) );
		write( fd, buf, size );

		free( buf );

		break;
		}
	case CMD_GETSTATE:
		{
		int	 handle;

		read( fd, &handle, sizeof(int) );

//		LP_DEBUG("Calling GetState(%d)...\n",  handle );

		int result = barh.GetState( handle );

//		LP_DEBUG("GetState() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_GETERROR:
		{
		int	 handle;

		read( fd, &handle, sizeof(int) );

//		LP_DEBUG("Calling GetError(%d)...\n", handle );

		int result = barh.GetError( handle );

//		LP_DEBUG("GetError() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_STARTUSINGCHANNEL:
		{
		int	handle, size;
		char*	buf;

		read( fd, &handle, sizeof( int) );
		read( fd, &size, sizeof( int) );
		if ( size )
		{
			buf = (char*)malloc( size );
			read( fd, buf, size );
		}
		LP_DEBUG("Calling StartUsingChannel(%d, %x)...\n", handle, size ? buf : NULL );

		int result = barh.StartUsingChannel( handle, size ? buf : NULL );

		LP_DEBUG("StartUsingChannel() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		if ( size )
		{
			free( buf );
		}

		break;
		}
	case CMD_STOPUSINGCHANNEL:
		{
		int	handle;

		read( fd, &handle, sizeof( int) );

		LP_DEBUG("Calling StopUsingChannel(%d)...\n", handle );

		int result = barh.StopUsingChannel( handle );

		LP_DEBUG("StopUsingChannel() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_GETAUTORECONNECT:
		{
		int	 handle;
		int	 autoRecconect;

		read( fd, &handle, sizeof(int) );

//		LP_DEBUG("Calling GetAutoReconnect(%d)...\n",  handle );

		bool result = barh.GetAutoReconnect( handle );

//		LP_DEBUG("GetAutoReconnect() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_SETAUTORECONNECT:
		{
		int	 handle, result = 0;
		int	 autoReconnect;

		read( fd, &handle, sizeof(int) );
		read( fd, &autoReconnect, sizeof(int) );

//		LP_DEBUG("Calling SetAutoReconnect(%d)...\n",  handle );

		barh.SetAutoReconnect( handle, autoReconnect );

//		LP_DEBUG("SetAutoReconnect() returned.\n");   

		write( fd, &result, sizeof(int) );

		break;
		}
	case CMD_ISPROFILEACTIVE:
		{
		int	 profile;
		int	 autoRecconect;

		read( fd, &profile, sizeof(int) );

//		LP_DEBUG("Calling IsProfileActive(%d)...\n",  profile );

		bool result = barh.IsProfileActive( profile );

//		LP_DEBUG("IsProfileActive() returned %d.\n", result);   

		write( fd, &result, sizeof(int) );

		break;
		}
	default:
		{
		BAS_ASSERT(false); //unrecognized command
		close( fd );
		}
	}

}

int main(int argc,char** argv)
{
	// server socket
	int			server_sockfd, client_sockfd;
	socklen_t		server_len, client_len;
	struct sockaddr_un	server_address, client_address;
	fd_set	readfds, testfds;

	if (argc<2)
	{
		LP_DEBUG("Use: %s <socket_name>\n", argv[0]);
		exit(1);
	}

	signal( SIGPIPE, SIG_IGN );
    
	unlink(argv[1]);

	server_sockfd = socket(AF_UNIX, SOCK_STREAM, 0);
	server_address.sun_family = AF_UNIX;
	strcpy(server_address.sun_path, argv[1]);
	server_len = sizeof(server_address);

	if (bind(server_sockfd, (struct sockaddr *)&server_address, server_len))
	{
		LP_DEBUG("BlueAclServer could not bind to the socket: %s\n", strerror(errno) );
		exit(1);
	}
	if (listen(server_sockfd, MAX_CLIENTS))
	{
		LP_DEBUG("BlueAclServer could not listen to the socket: %s\n", strerror(errno));
		exit(1);
	}

	FD_ZERO(&readfds);
	FD_SET(server_sockfd, &readfds);

	while(serverbusy)
	{
		int result;
		int fd;

	        testfds = readfds;
        	result = select(FD_SETSIZE, &testfds, (fd_set *)0, (fd_set *)0, (struct timeval *) 0);

		for(fd = 0; fd < FD_SETSIZE; fd++)
		{
			if(FD_ISSET(fd,&testfds))
			{
				if(fd == server_sockfd)
				{
					//server received new connection request
					client_len = sizeof(client_address);
					client_sockfd = accept(server_sockfd, (struct sockaddr *)&client_address, (socklen_t*)&client_len);
					FD_SET(client_sockfd, &readfds);
					LP_DEBUG("BlueAclServer: client added on fd = %d!\n", client_sockfd);
				}
				else
				{
					int nread;

					ioctl(fd, FIONREAD, &nread);

					if(nread == 0)
					{
						close(fd);
						FD_CLR(fd, &readfds);
						LP_DEBUG("BlueAclServer: client removed on fd = %d!\n", fd);
					}
					else 
					{
						requesthandler(fd);
					}
				}
			}
                }

	}

	return 0;
}
