#include <dirent.h>
#include <ctype.h>

#include "bluetoothchannel.h"
#define MAX_CONNECTIONS	1

//#define DEBUG_BLUETOOTHCHANNEL

#ifdef DEBUG_BLUETOOTHCHANNEL
  #define LP_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LP_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LP_DEBUG(msg...)     do { } while (0)
  #define LP_DEBUGT(t, msg...) do { } while (0)
#endif

//if you want to have asserts on 0 -> 1
#if 0
  #include <assert.h>
  #define BT_ASSERT(x) do{ assert(x); } while(0)
#else
  #define BT_ASSERT(x) do { } while (0)
#endif

BluetoothChannel::BluetoothChannel()
{
	sprintf( btAddr, "" );
	btChannel = -1;
	sprintf( btTTY, "" );
	btDev = -1;
	btFd = -1;
}

int BluetoothChannel::Initialize(char* aAddr, int aChannel)
{
	//initialize device info
	strcpy( btAddr, aAddr );
	btChannel = aChannel;
	sprintf( btTTY, "" );
	btDev = -1;
	btFd = -1;
}

int BluetoothChannel::Open()
{

	struct sockaddr_rc remote_addr, local_addr;
	bdaddr_t bdaddr;
	int s;

	str2ba(btAddr, &bdaddr);
	if ((s = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) {
		LP_DEBUG("Can't create socket. %s (%d)\n", strerror(errno), errno);
		return -1;
	}

	memset(&local_addr, 0, sizeof(local_addr));
	local_addr.rc_family = AF_BLUETOOTH;
	bacpy(&local_addr.rc_bdaddr, BDADDR_ANY);
	if (bind(s, (struct sockaddr *)&local_addr, sizeof(local_addr)) < 0) {
		LP_DEBUG("Can't bind socket. %s (%d)\n", strerror(errno), errno);
		close(s);
		return -1;
	}

	memset(&remote_addr, 0, sizeof(remote_addr));
	remote_addr.rc_family = AF_BLUETOOTH;
	bacpy(&remote_addr.rc_bdaddr, &bdaddr);
	remote_addr.rc_channel = btChannel;
	if (connect(s, (struct sockaddr *)&remote_addr, sizeof(remote_addr)) < 0) {
		LP_DEBUG("Can't connect. %s (%d)\n", strerror(errno), errno);
		close(s);
		return -1;
	}
	else
	{
		if ( fcntl( s, F_SETFL, fcntl( s, F_GETFL ) | O_NONBLOCK ) < 0 )
			LP_DEBUG("BlueToothChannel::Send(): cannot set to non-blocking mode\n");
		btFd = s;
		LP_DEBUG("RFCOMM channel connected, fd = %d\n", btFd);
	}

	return 0;
}
int BluetoothChannel::Listen(int aChannel)
{

	bdaddr_t		local;
	bdaddr_t		bdaddr;
	struct sockaddr_rc 	addr;
	socklen_t		socklen;
	int 			s;

	hci_devba(0, &local);
	str2ba(btAddr, &bdaddr);

	if ((btFd = socket(PF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM)) < 0) 
	{
		LP_DEBUG("Can't create socket. %s (%d)\n", strerror(errno), errno);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, &local);
	addr.rc_channel = aChannel;
	if (bind(btFd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
	{
		LP_DEBUG("Can't bind socket. %s (%d)\n", strerror(errno), errno);
		close(btFd);
		return -1;
	}

	if (listen(btFd, MAX_CONNECTIONS)<0)
	{
		LP_DEBUG("Can't listen socket. %s (%d)\n", strerror(errno), errno);
		close(btFd);
		return -1;
	}

	memset(&addr, 0, sizeof(addr));
	addr.rc_family = AF_BLUETOOTH;
	bacpy(&addr.rc_bdaddr, &bdaddr);
	addr.rc_channel = aChannel;
	if ((s = accept(btFd, (struct sockaddr *)&addr, &(socklen = sizeof(addr)))) < 0 )
	{
		LP_DEBUG("Can't accept socket. %s (%d)\n", strerror(errno), errno);
		close(btFd);
		return -1;
	}
	else
	{
		close(btFd);
		if ( fcntl( s, F_SETFL, fcntl( s, F_GETFL ) | O_NONBLOCK ) < 0 )
			LP_DEBUG("BlueToothChannel::Listen(): cannot set to non-blocking mode\n");
		btFd = s;
		LP_DEBUG("RFCOMM channel accepted\n");
	}
	
	return 0;
}
int BluetoothChannel::Close()
{
	int result = 0;

	result = ( btFd < 0 ) ? 0 : close(btFd);
	result += ReleaseTTY();

	if (result < 0 )
	{
		LP_DEBUG("BlueToothChannel::Close(): Closing channel failed: %s (%d)\n", strerror(errno), errno);
	}

	btFd = -1;

	return result;
}

int BluetoothChannel::Send(char* aBuffer, int aSize)
{

	if ( btFd < 0 )
	{
		LP_DEBUG("BlueToothChannel::Send(): Sending failed due to invalid file descriptor.\n");
		return -1;
	}
	return write(btFd, aBuffer, aSize);

}

int BluetoothChannel::Receive(char* aBuffer, int* aBufferLength)
{

	if ( btFd < 0 )
	{
		LP_DEBUG("BlueToothChannel::Receive(): Receiving failed due to invalid file descriptor.\n");
		return -1;
	}


	int		n, result;

	n = read(btFd, aBuffer, *aBufferLength );
	if ( n < 0 )
	{
		n = 0;
		if ( errno == EAGAIN )
		{
			result = 0;
		}
		else
		{
			LP_DEBUG("BlueToochChannel::Receive(): Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
			result = -1;
		}
	}
	else
		result = 0;

	*aBufferLength = n;

	return result;
/*
	if ( btFd < 0 )
	{
		LP_DEBUG("BlueToothChannel::Receive(): Receiving failed due to invalid file descriptor.\n");
		return -1;
	}


	int		nfds, n, result;
	struct timeval	timeout;
	fd_set		testfds;

	FD_ZERO(&testfds);
	FD_SET(btFd, &testfds);
    
	//do not wait
	timeout.tv_sec = 0;
	timeout.tv_usec = 0;
     
	//check if anything to read 
	nfds = select(FD_SETSIZE, &testfds, (fd_set *)0,(fd_set *)0, &timeout);

	n = 0;

	if ( nfds > 0 )
	{
		if(FD_ISSET(btFd,&testfds))
		{
			long nread = 1;
			//ioctl(btFd, FIONREAD, &nread); //todo: find a way to check if the read on the socket will be non-blocking

			if ( nread )
			{
				n = read(btFd, aBuffer, *aBufferLength - 1);
				if ( n < 0 )
				{
					LP_DEBUG("BlueToochChannel::Receive(): Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
					n = 0;
					result = -1;
				}
				else
				{
					result = 0;
				}
			}
			else
			{
				result = 0;
			}
		}
		else
		{
			BT_ASSERT(false);
		}
		*aBufferLength = n;
	}
	else if ( nfds == 0 )
	{
		result = 0;
		*aBufferLength = 0;
	}
	else
	{
		result = -1;
		*aBufferLength = 0;
	}


	return result;
*/
}

int	BluetoothChannel::Send(const unsigned char* aBuffer, int aBufferLength)
{
//	BT_ASSERT( devFd == -1 );
	return write(btFd, aBuffer, aBufferLength);
}

int	BluetoothChannel::Receive(unsigned char* aBuffer, int aBufferLength, int aTimeout)
{
//	BT_ASSERT( devFd == -1 );
	int		nfds, n, result;
	struct timeval	timeout;
	fd_set		testfds;

	FD_ZERO(&testfds);
	FD_SET(btFd, &testfds);
    
	//do not wait
	timeout.tv_sec = aTimeout / 1000;
	timeout.tv_usec = (aTimeout - timeout.tv_sec * 1000) * 1000;
     
	//check if anything to read 
	nfds = select(FD_SETSIZE, &testfds, (fd_set *)0,(fd_set *)0, &timeout);

	n = 0;

	if ( nfds > 0 )
	{
		if(FD_ISSET(btFd,&testfds))
		{
			long nread = 1;
			//ioctl(btFd, FIONREAD, &nread); //todo: find a way to check if the read on the socket will be non-blocking

			if ( nread )
			{
				n = read(btFd, aBuffer, aBufferLength);
				if ( n < 0 )
				{
					LP_DEBUG("Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
					n = 0;
					result = -1;
				}
				else
				{
					result = 0;
				}
			}
			else
			{
				result = 0;
			}
		}
		else
		{
			//this should never happen: we are waiting for data on btFd descriptor
			BT_ASSERT(false);
		}
	}
	else if ( nfds == 0 )
	{
		result = 0;
	}
	else
	{
		LP_DEBUG("Select on the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
		result = -1;
	}
	return result;
}

int	BluetoothChannel::SendBinary(const unsigned char* aBuffer, int aBufferLength)
{
//	BT_ASSERT( devFd == -1 );
	return write(btFd, aBuffer, aBufferLength);
}

int	BluetoothChannel::ReceiveBinary(char* aBuffer, int aBufferLength, int aTimeout)
{



//	BT_ASSERT( devFd == -1 );


  int		nfds, n, result;
	struct timeval	timeout;
	fd_set		testfds;

	FD_ZERO(&testfds);
	FD_SET(btFd, &testfds);

//printf("btFd = %d\n",btFd);


	//do not wait
	timeout.tv_sec = aTimeout/1000;
	timeout.tv_usec = (aTimeout%1000)*1000;
     
	//check if anything to read 
	nfds = select(btFd+1/*FD_SETSIZE*/, &testfds, (fd_set *)0,(fd_set *)0, &timeout);

//printf("nfds = %d\n",nfds);

	n = 0;

	if ( nfds > 0 )
	{
		if(FD_ISSET(btFd,&testfds))
		{
//printf("FD_ISSET(btFd,&testfds)\n");
			n = read(btFd, aBuffer, aBufferLength);
//printf("nread = %d\n",n);
			if ( n < 0 )
			{
				LP_DEBUG("Reading form the %s channel %d failed: %s (%d)\n", btAddr, btChannel, strerror(errno), errno);
				result = 0;
			}
			else
			{
				result = n;
			}
		}
		else
		{
			//this should never happen: we are waiting for data on btFd descriptor
			BT_ASSERT(false);
		}
	}
	else if ( nfds == 0 )
	{
		result = 0;
	}
	else
	{
		result = 0;
	}

	return result;
}

int BluetoothChannel::CreateTTY(int dev_id)
{
	int ctl, ntry = 3;
	struct rfcomm_dev_req req;
	struct stat st;


	if ( dev_id == btDev )
	{
		LP_DEBUG("BlueToothChanneli:CreateTTY(%d): reusing existing device.\n", dev_id);
		return 0;
	}

	memset(&req, 0, sizeof(req));
	req.dev_id = dev_id;
	bacpy(&req.src, BDADDR_ANY);
	str2ba(btAddr, &req.dst);
	req.channel = btChannel;

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
		LP_DEBUG("BlueToothChannel::CreateTTY(%d): Can't open RFCOMM control socket\n", dev_id);
		return -1;
	}

	if ( ioctl(ctl, RFCOMMCREATEDEV, &req) < 0 )
	{
		LP_DEBUG("BlueToothChannel::CreateTTY(%d): Can't create device\n", dev_id);
		close( ctl );
		return -1;
	}

	sprintf(btTTY, "/dev/rfcomm%d", dev_id);
	while (stat(btTTY, &st) < 0) {
		sprintf(btTTY, "/dev/bluetooth/rfcomm/%d", dev_id);
		if (stat(btTTY, &st) < 0) {
			sprintf(btTTY, "/dev/rfcomm%d", dev_id);
			if (ntry--) {
				sleep(1);
				continue;
			}

			memset(&req, 0, sizeof(req));
			req.dev_id = dev_id;
			ioctl(ctl, RFCOMMRELEASEDEV, &req);
			sprintf(btTTY, "");
			close( ctl );
			return -1;
		}
	}

	btDev = dev_id;
	close( ctl );

	return 0;
}

char* BluetoothChannel::GetTTY()
{
	return btTTY;
}
bool BluetoothChannel::IsTTYBusy()
{
	static bool rfcomm_state[] = {
		true,//"unknown",
		true,//"connected",
		false,//"clean",
		true,//"bound",
		true,//"listening",
		true,//"connecting",
		true,//"connecting",
		true,//"config",
		true,//"disconnecting",
		false//"closed"
	};

	int ctl, i;
	bool result;
	struct rfcomm_dev_list_req *dl;
	struct rfcomm_dev_info *di;

	if ( btDev < 0 )
	{
		LP_DEBUG("WARNING: BlueToothChannel::IsTTYBusy(): device does not exist.\n");
		return -1;
	}

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
		LP_DEBUG("BlueToothChannel::IsTTYBusy(): Can't open RFCOMM control socket");
		return -1;
	}

	dl = (rfcomm_dev_list_req*)malloc(sizeof(*dl) + RFCOMM_MAX_DEV * sizeof(*di));
	if (!dl) {
		LP_DEBUG("BlueToothChannel::IsTTYBusy(): Can't allocate memory");
		close( ctl );
		return -1;
	}

	dl->dev_num = RFCOMM_MAX_DEV;
	di = dl->dev_info;
	if (ioctl(ctl, RFCOMMGETDEVLIST, (void *) dl) < 0) {
		LP_DEBUG("BlueToothChannel::IsTTYBusy(): Can't get device list");
		close( ctl );
		return -1;
	}

	printf("dev_num = %d\n", dl->dev_num);
	result = false;
	for (i = 0; i < dl->dev_num; i++)
	{
		if ( di->id == btDev )
		{
			printf("dev %d, state = %d\n", di->id, (di + i)->state);
			result = rfcomm_state[(di + i)->state];
			break;
		}
	}

	free( dl );
	close( ctl );

	return result;
}

int BluetoothChannel::ReleaseTTY()
{
	int ctl;
	struct rfcomm_dev_req req;

	if ( btDev < 0 )
	{
		LP_DEBUG("WARNING: BlueToothChannel::ReleaseTTY(): device does not exist.\n");
		return 0;
	}

	memset(&req, 0, sizeof(req));
	req.dev_id = btDev;

	if ((ctl = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM)) < 0 ) {
		LP_DEBUG("BlueToothChannel::ReleaseTTY(): Can't open RFCOMM control socket");
		return -1;
	}

	if ( ioctl(ctl, RFCOMMRELEASEDEV, &req)  < 0 )
	{
		LP_DEBUG("BlueToothChannel::ReleaseTTY(): Can't release device");
		close( ctl );
		return -1;
	}

	btDev = -1;
	sprintf(btTTY, "");
	close( ctl );

	return 0;
}

char* BluetoothChannel::GetAddress()
{
	return btAddr;
}

int BluetoothChannel::GetChannel()
{
	return btChannel;
}
