#include <pthread.h>

#define DEFAULT_STACK_SIZE (16*1024) //default stacj size for the request handler
#define MUTEX_WAIT_RELEASE_TIME 100 //how much time to wait on mutex to be released
#define MUTEX_WAIT_DESTROY_TIME 1000 //how much time to wait on mutex to be released

class RequestPool
{
public:
	enum
	{
		RP_DESTROY_SYNC,
		RP_DESTROY_ASYNC,
		RP_DESTROY_NOW
	} DESTRUCTION_TYPE;
	enum
	{
		RP_ADD_HEAD,	
		RP_ADD_TAIL
	} ADD_TYPE;

	RequestPool();

	int Initialize( int aStackSize );

	int Add( void (*function)(void *),  void * argument, int direction ); //add one request

	int Add2( void (*function1)(void *),  void * argument1, void (*function2)(void *), void * argument2, int direction ); //ass 2 requests

	int Count(void); //the number of requests pending

	void Empty(void); //remove all the requests

	int Destroy( int aNow );


private:
	/*
	 * request_item_t	- Structure representing a request.
	 */
	typedef struct request_item
	{
	    struct request_item * next;
	    void 		(*function)(void *);
	    void		*argument;

	} request_item_t;

	/*
	 * request_pool_t	- Structure describing a request pool.
	 */
	typedef struct request_pool
	{
	    pthread_mutex_t	mutex;
	    pthread_cond_t	cond;

	    request_item_t	*head, *tail;	// head and tail of work queue.

	    int			quit;
	    int			valid;
	    int			destroyed;


	} request_pool_t;

	static void* ThreadFunction(void *aPtr);	// adaptor pThread to C++
	void RequestHandler();				//handle incoming open/send/receive/close requests
	static void ThreadFreePool(void* aPtr); //clean up routine adaptor
	void FreePool(); //cleaning up routine

	pthread_t	handlerThread;
	pthread_attr_t	handlerThreadAttr;
	request_pool_t	requestPool;
	int		requestCount;
	void*		pendingArgument;
};
